"""Tests for quantization functionality (Task3.1)."""

from __future__ import annotations

import tempfile
from pathlib import Path

import numpy as np
import pytest

from sagellm_compression.quantization import (
    Calibrator,
    Exporter,
    FP8Config,
    Quantizer,
    W4A16Config,
    W8A8Config,
)
from sagellm_compression.quantization.configs import CalibrationMethod, QuantizationType


class TestQuantizationConfigs:
    """Test quantization configuration classes."""

    def test_w8a8_config_default(self) -> None:
        """Test W8A8 config with default parameters."""
        config = W8A8Config()
        assert config.quantization_type == QuantizationType.W8A8
        assert config.per_channel is True
        assert config.symmetric is True
        assert config.quantize_weights is True
        assert config.quantize_activations is True

    def test_w4a16_config_default(self) -> None:
        """Test W4A16 config with default parameters."""
        config = W4A16Config()
        assert config.quantization_type == QuantizationType.W4A16
        assert config.per_channel is True
        assert config.symmetric is False  # W4A16 uses asymmetric
        assert config.quantize_weights is True
        assert config.quantize_activations is False  # Activations stay in FP16
        assert config.group_size == 128

    def test_fp8_config_default(self) -> None:
        """Test FP8 config with default parameters."""
        config = FP8Config()
        assert config.quantization_type == QuantizationType.FP8
        assert config.per_channel is True
        assert config.symmetric is True
        assert config.fp8_format == "e4m3"

    def test_custom_calibration_method(self) -> None:
        """Test custom calibration method."""
        config = W8A8Config(calibration_method=CalibrationMethod.PERCENTILE)
        assert config.calibration_method == CalibrationMethod.PERCENTILE

    def test_custom_num_samples(self) -> None:
        """Test custom number of calibration samples."""
        config = W8A8Config(num_calibration_samples=256)
        assert config.num_calibration_samples == 256

    def test_w4a16_custom_group_size(self) -> None:
        """Test W4A16 with custom group size."""
        config = W4A16Config(group_size=64)
        assert config.group_size == 64


class TestCalibrator:
    """Test Calibrator class."""

    def test_calibrator_initialization(self) -> None:
        """Test calibrator initialization."""
        config = W8A8Config()
        calibrator = Calibrator(config)
        assert calibrator.config == config
        assert len(calibrator.samples) == 0
        assert calibrator.params is None

    def test_collect_samples(self) -> None:
        """Test collecting calibration samples."""
        config = W8A8Config(num_calibration_samples=10)
        calibrator = Calibrator(config)

        # Collect some samples
        for i in range(5):
            sample = np.random.randn(32, 64).astype(np.float32)
            calibrator.collect(sample)

        assert len(calibrator.samples) == 5

    def test_compute_params_minmax(self) -> None:
        """Test computing quantization parameters with minmax calibration."""
        config = W8A8Config(calibration_method=CalibrationMethod.MINMAX)
        calibrator = Calibrator(config)

        # Collect samples with known range
        samples = [
            np.array([[-2.0, -1.0], [0.0, 1.0]], dtype=np.float32),
            np.array([[-1.5, 0.5], [1.5, 2.0]], dtype=np.float32),
        ]
        for sample in samples:
            calibrator.collect(sample)

        params = calibrator.compute_params(bits=8)
        assert params is not None
        assert params.min_val == pytest.approx(-2.0, abs=1e-5)
        assert params.max_val == pytest.approx(2.0, abs=1e-5)

    def test_compute_params_symmetric(self) -> None:
        """Test symmetric quantization (zero_point = 0)."""
        config = W8A8Config(symmetric=True)
        calibrator = Calibrator(config)

        sample = np.array([[-1.0, 0.0, 1.0]], dtype=np.float32)
        calibrator.collect(sample)

        params = calibrator.compute_params(bits=8)
        assert params.zero_point == 0  # Symmetric

    def test_compute_params_no_samples_raises(self) -> None:
        """Test that computing params without samples raises error."""
        config = W8A8Config()
        calibrator = Calibrator(config)

        with pytest.raises(ValueError, match="No samples collected"):
            calibrator.compute_params()

    def test_calibrator_reset(self) -> None:
        """Test resetting calibrator state."""
        config = W8A8Config()
        calibrator = Calibrator(config)

        sample = np.random.randn(10, 10).astype(np.float32)
        calibrator.collect(sample)
        calibrator.compute_params()

        calibrator.reset()
        assert len(calibrator.samples) == 0
        assert calibrator.params is None


class TestQuantizer:
    """Test Quantizer class."""

    def test_quantizer_initialization(self) -> None:
        """Test quantizer initialization."""
        config = W8A8Config()
        quantizer = Quantizer(config)
        assert quantizer.config == config

    def test_quantize_tensor_w8a8(self) -> None:
        """Test quantizing a single tensor with W8A8."""
        config = W8A8Config()
        quantizer = Quantizer(config)

        # Create a sample tensor
        tensor = np.array([[-1.0, 0.0, 1.0], [2.0, 3.0, 4.0]], dtype=np.float32)

        # Compute quantization params
        calibrator = Calibrator(config)
        calibrator.collect(tensor)
        params = calibrator.compute_params(bits=8)

        # Quantize
        quantized = quantizer.quantize_tensor(tensor, params, bits=8)
        assert quantized.data.dtype == np.int8
        assert quantized.original_shape == tensor.shape

        # Dequantize and check approximate reconstruction
        dequantized = quantized.dequantize()
        assert dequantized.shape == tensor.shape
        # Allow some quantization error
        assert np.allclose(dequantized, tensor, atol=0.1)

    def test_quantize_model_weights(self) -> None:
        """Test quantizing model weights."""
        config = W8A8Config()
        quantizer = Quantizer(config)

        # Simulate model weights
        model_weights = {
            "layer1.weight": np.random.randn(64, 128).astype(np.float32),
            "layer2.weight": np.random.randn(128, 256).astype(np.float32),
        }

        quantized_model = quantizer.quantize(model_weights)
        assert len(quantized_model.weights) == 2
        assert "layer1.weight" in quantized_model.weights
        assert "layer2.weight" in quantized_model.weights
        assert quantized_model.metadata["num_layers"] == 2

    def test_quantize_w4a16(self) -> None:
        """Test W4A16 quantization."""
        config = W4A16Config(group_size=64)
        quantizer = Quantizer(config)

        model_weights = {
            "weight": np.random.randn(128, 128).astype(np.float32),
        }

        quantized_model = quantizer.quantize(model_weights)
        assert len(quantized_model.weights) == 1
        # W4A16 uses 4-bit weights
        weight = quantized_model.weights["weight"]
        assert weight.data.dtype == np.int8  # Stored as int8


class TestExporter:
    """Test Exporter class."""

    def test_exporter_initialization(self) -> None:
        """Test exporter initialization."""
        exporter = Exporter()
        assert exporter is not None

    def test_export_and_load_pickle(self) -> None:
        """Test exporting and loading in pickle format."""
        config = W8A8Config()
        quantizer = Quantizer(config)

        model_weights = {
            "layer.weight": np.random.randn(32, 64).astype(np.float32),
        }
        quantized_model = quantizer.quantize(model_weights)

        exporter = Exporter()

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = Path(tmpdir) / "model.pkl"

            # Export
            exporter.export(quantized_model, output_path, format="pickle")
            assert output_path.exists()

            # Load
            loaded_model = exporter.load(output_path, format="pickle")
            assert len(loaded_model.weights) == len(quantized_model.weights)
            assert loaded_model.config.quantization_type == config.quantization_type

    def test_export_metadata(self) -> None:
        """Test exporting model metadata to JSON."""
        config = W8A8Config()
        quantizer = Quantizer(config)

        model_weights = {
            "layer.weight": np.random.randn(16, 32).astype(np.float32),
        }
        quantized_model = quantizer.quantize(model_weights)

        exporter = Exporter()

        with tempfile.TemporaryDirectory() as tmpdir:
            metadata_path = Path(tmpdir) / "metadata.json"

            exporter.export_metadata(quantized_model, metadata_path)
            assert metadata_path.exists()

            # Check content
            import json

            with open(metadata_path) as f:
                metadata = json.load(f)

            assert metadata["quantization_type"] == "w8a8"
            assert metadata["num_layers"] == 1
            assert "layer.weight" in metadata["layer_names"]

    def test_load_nonexistent_file_raises(self) -> None:
        """Test loading non-existent file raises error."""
        exporter = Exporter()

        with pytest.raises(FileNotFoundError):
            exporter.load("/nonexistent/path/model.pkl")


class TestEndToEnd:
    """End-to-end integration tests."""

    def test_full_quantization_pipeline_w8a8(self) -> None:
        """Test full quantization pipeline with W8A8."""
        # 1. Create config
        config = W8A8Config()

        # 2. Quantize model
        quantizer = Quantizer(config)
        model_weights = {
            "fc1.weight": np.random.randn(128, 256).astype(np.float32),
            "fc2.weight": np.random.randn(256, 512).astype(np.float32),
        }
        quantized_model = quantizer.quantize(model_weights)

        # 3. Export
        exporter = Exporter()
        with tempfile.TemporaryDirectory() as tmpdir:
            model_path = Path(tmpdir) / "model_int8.pkl"
            exporter.export(quantized_model, model_path)

            # 4. Load
            loaded_model = exporter.load(model_path)

            # 5. Verify
            assert len(loaded_model.weights) == 2
            assert set(loaded_model.weights.keys()) == {"fc1.weight", "fc2.weight"}

            # Dequantize and check reconstruction quality
            for layer_name, original_weight in model_weights.items():
                quantized_weight = loaded_model.weights[layer_name]
                dequantized = quantized_weight.dequantize()
                # Check shape matches
                assert dequantized.shape == original_weight.shape
                # Check quantization introduces bounded error
                error = np.abs(dequantized - original_weight).mean()
                assert error < 0.5  # Reasonable quantization error for INT8

    def test_full_quantization_pipeline_w4a16(self) -> None:
        """Test full quantization pipeline with W4A16."""
        config = W4A16Config(group_size=64)
        quantizer = Quantizer(config)

        model_weights = {
            "weight": np.random.randn(64, 64).astype(np.float32),
        }
        quantized_model = quantizer.quantize(model_weights)

        exporter = Exporter()
        with tempfile.TemporaryDirectory() as tmpdir:
            model_path = Path(tmpdir) / "model_int4.pkl"
            exporter.export(quantized_model, model_path)
            loaded_model = exporter.load(model_path)

            assert len(loaded_model.weights) == 1
            assert loaded_model.config.quantization_type == QuantizationType.W4A16


def test_quantization_placeholder() -> None:
    """Placeholder test to ensure test suite runs."""
    assert True, "Quantization module placeholder test"



"""Tests for speculative decoding functionality (Task3.3)."""

from __future__ import annotations

from dataclasses import dataclass

import pytest

from sagellm_compression.speculative import (
    DraftModel,
    SpeculativeController,
    VerificationResult,
    Verifier,
)


@dataclass
class SequentialDraftModel(DraftModel):
    """Draft model that generates consecutive ids."""

    def propose_tokens(self, context_ids: list[int], n_tokens: int) -> list[int]:
        start = context_ids[-1] if context_ids else 0
        return [start + i + 1 for i in range(n_tokens)]


@dataclass
class PrefixVerifier(Verifier):
    """Verifier that accepts a fixed prefix length."""

    accept_len: int

    def verify_tokens(self, context_ids: list[int], draft_tokens: list[int]) -> VerificationResult:
        accepted = draft_tokens[: self.accept_len]
        if self.accept_len < len(draft_tokens):
            return VerificationResult(
                accepted_tokens=accepted,
                rejected_token=draft_tokens[self.accept_len] + 100,
            )
        return VerificationResult(accepted_tokens=accepted, rejected_token=None)


@dataclass
class EmptyDraftModel(DraftModel):
    """Invalid draft model for fail-fast tests."""

    def propose_tokens(self, context_ids: list[int], n_tokens: int) -> list[int]:
        return []


@dataclass
class InvalidVerifier(Verifier):
    """Invalid verifier for fail-fast tests."""

    def verify_tokens(self, context_ids: list[int], draft_tokens: list[int]) -> VerificationResult:
        wrong_prefix = list(reversed(draft_tokens))
        return VerificationResult(accepted_tokens=wrong_prefix, rejected_token=None)


@pytest.mark.speculative
class TestSpeculativeController:
    """Speculative draft-verify controller tests."""

    def test_generate_all_accepted(self) -> None:
        controller = SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=PrefixVerifier(accept_len=8),
            n_speculative_tokens=4,
        )

        result = controller.generate(input_ids=[10], max_new_tokens=8)

        assert result.generated_tokens == [11, 12, 13, 14, 15, 16, 17, 18]
        assert result.metrics.accepted_tokens_total == 8
        assert result.metrics.rejected_tokens_total == 0
        assert result.metrics.accept_rate == 1.0

    def test_generate_partial_accept_with_fallback(self) -> None:
        controller = SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=PrefixVerifier(accept_len=2),
            n_speculative_tokens=4,
        )

        result = controller.generate(input_ids=[1], max_new_tokens=6)

        assert result.generated_tokens == [2, 3, 104, 105, 106, 207]
        assert result.metrics.accepted_tokens_total == 4
        assert result.metrics.rejected_tokens_total == 2
        assert result.metrics.draft_tokens_total == 7
        assert result.metrics.accept_rate == pytest.approx(4 / 7, abs=1e-9)

    def test_fail_fast_empty_draft(self) -> None:
        controller = SpeculativeController(
            draft_model=EmptyDraftModel(),
            verifier=PrefixVerifier(accept_len=1),
            n_speculative_tokens=3,
        )

        with pytest.raises(ValueError, match="empty token list"):
            controller.generate(input_ids=[1], max_new_tokens=3)

    def test_fail_fast_invalid_accept_prefix(self) -> None:
        controller = SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=InvalidVerifier(),
            n_speculative_tokens=3,
        )

        with pytest.raises(ValueError, match="prefix"):
            controller.generate(input_ids=[1], max_new_tokens=3)

    def test_fail_fast_all_accepted_with_rejected_token(self) -> None:
        class BadVerifier(Verifier):
            def verify_tokens(
                self,
                context_ids: list[int],
                draft_tokens: list[int],
            ) -> VerificationResult:
                return VerificationResult(accepted_tokens=draft_tokens, rejected_token=42)

        controller = SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=BadVerifier(),
            n_speculative_tokens=2,
        )

        with pytest.raises(ValueError, match="must be None"):
            controller.generate(input_ids=[1], max_new_tokens=2)

    def test_metrics_serialization(self) -> None:
        controller = SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=PrefixVerifier(accept_len=1),
            n_speculative_tokens=2,
        )

        result = controller.generate(input_ids=[1], max_new_tokens=3)
        payload = result.to_dict()

        assert "metrics" in payload
        assert payload["metrics"]["accept_rate"] <= 1.0
        assert payload["metrics"]["spec_accept_rate"] == payload["metrics"]["accept_rate"]
        assert payload["metrics"]["estimated_speedup"] >= 1.0


@pytest.mark.speculative
def test_invalid_controller_arguments() -> None:
    """Controller should validate constructor arguments."""
    with pytest.raises(ValueError, match="n_speculative_tokens"):
        SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=PrefixVerifier(accept_len=1),
            n_speculative_tokens=0,
        )

    with pytest.raises(ValueError, match="max_iterations"):
        SpeculativeController(
            draft_model=SequentialDraftModel(),
            verifier=PrefixVerifier(accept_len=1),
            n_speculative_tokens=1,
            max_iterations=0,
        )


@pytest.mark.speculative
def test_invalid_generate_arguments() -> None:
    """Generate should validate runtime arguments."""
    controller = SpeculativeController(
        draft_model=SequentialDraftModel(),
        verifier=PrefixVerifier(accept_len=1),
        n_speculative_tokens=2,
    )

    with pytest.raises(ValueError, match="max_new_tokens"):
        controller.generate(input_ids=[1], max_new_tokens=0)



"""sagellm-compression: Model Compression & Acceleration Module for sageLLM."""

from __future__ import annotations

from sagellm_compression._version import __version__
from sagellm_compression.quantization import (
    Calibrator,
    Exporter,
    FP8Config,
    QuantizationConfig,
    Quantizer,
    W4A16Config,
    W8A8Config,
)
from sagellm_compression.speculative import (
    DraftModel,
    SpeculativeController,
    SpeculativeMetrics,
    SpeculativeResult,
    VerificationResult,
    Verifier,
)

# Public API
__all__ = [
    "__version__",
    # Quantization (Task 3.1)
    "Calibrator",
    "Quantizer",
    "Exporter",
    "QuantizationConfig",
    "W8A8Config",
    "W4A16Config",
    "FP8Config",
    # Speculative decoding (Task 3.3)
    "DraftModel",
    "Verifier",
    "VerificationResult",
    "SpeculativeController",
    "SpeculativeResult",
    "SpeculativeMetrics",
    # Future: Sparsity, KernelFusion
]
